from ngsolve import *
from netgen.geom2d import unit_square
from math import pi

order = 2
epsilon = 1 # 1=SIPG
eliminate_up = 1 # 1=Eliminate both 0=Eliminate u only

exact_uu = CoefficientFunction((sin(pi*x)*sin(pi*y), \
                                cos(pi*x)*cos(pi*y)))
exact_p  = CoefficientFunction(sin(pi*x)*cos(pi*y))
sourcexy = CoefficientFunction((2.0*pi*pi*sin(pi*x)*sin(pi*y) + pi*cos(pi*x)*cos(pi*y), \
                                2.0*pi*pi*cos(pi*x)*cos(pi*y) - pi*sin(pi*x)*sin(pi*y)))

mesh = Mesh(unit_square.GenerateMesh(maxh=0.05))
print ("boundary labels: ", mesh.GetBoundaries())

# Function spaces
V    = VectorL2(mesh, order = order)
Q    = L2(mesh, order = order-1)
Vbar = FacetFESpace(mesh, order = order, dirichlet= "bottom|right|top|left")
Qbar = FacetFESpace(mesh, order = order)  
X    = FESpace([V, Vbar, Vbar, Q, Qbar])

uu, ubarx, ubary, p, pbar = X.TrialFunction()
vv, vbarx, vbary, q, qbar = X.TestFunction()
ubar = CoefficientFunction((ubarx, ubary))
vbar = CoefficientFunction((vbarx, vbary))

n = specialcf.normal(mesh.dim)
h = specialcf.mesh_size

# Penalty parameter
alpha = 6.0*order**2/h

# Bi-linear form
a_stokes = BilinearForm(X, eliminate_internal = True)

ah_element = InnerProduct(grad(uu), grad(vv))
ah_facet   = alpha*(uu-ubar)*(vv-vbar) \
             - epsilon*InnerProduct(uu-ubar, grad(vv)*n) \
             - InnerProduct(vv-vbar, grad(uu)*n)
bh_element = - p*Trace(grad(vv)) - q*Trace(grad(uu))
bh_facet   = (vv-vbar)*n*pbar + (uu-ubar)*n*qbar

a_stokes += SymbolicBFI(ah_element + bh_element)
a_stokes += SymbolicBFI(ah_facet + bh_facet, element_boundary=True)

# Linear form
f_stokes = LinearForm(X)
f_stokes += SymbolicLFI(sourcexy*vv)

# For solver
c = Preconditioner(type="direct", bf=a_stokes, inverse = "umfpack")

# gridfunction for the solution
gfu = GridFunction(X)

def CalcL2Error(sol):
    err_u = sqrt(Integrate((sol.components[0]-exact_uu)**2, mesh))
    pressure_mean = Integrate(sol.components[3], mesh)
    err_p = sqrt(Integrate((sol.components[3]-pressure_mean-exact_p)**2, mesh))
    err_div = sqrt(Integrate(Trace(grad(sol.components[0]))**2, mesh))
    return (err_u, err_p, err_div)

def SolveBVP():

    X.Update()
    print("ndof:", X.ndof)
    gfu.Update()
    start = X.components[0].ndof + X.components[1].ndof + X.components[2].ndof
    end = start + X.components[3].ndof

    for ii in range(start, end):
       if (eliminate_up==0):  
          X.SetCouplingType(ii, COUPLING_TYPE.INTERFACE_DOF)

    # boundary condition
    gfu.components[1].Set(exact_uu[0], BND)
    gfu.components[2].Set(exact_uu[1], BND)
    
    a_stokes.Assemble()
    f_stokes.Assemble()

    # solve Stokes problem:
    BVP(bf=a_stokes, lf=f_stokes, gf=gfu, pre=c, maxsteps=10000).Do()
    err_u, err_p, err_div = CalcL2Error(gfu)
    print ("err_u, err_p, err_div")
    print (err_u, err_p, err_div)

    # for visualization
    velocity = CoefficientFunction (gfu.components[0])
    Draw (velocity, mesh, "velocity", sd=3)

    pressure = CoefficientFunction (gfu.components[3])
    Draw (pressure, mesh, "pressure", sd=3)

SolveBVP()
